{
  Batch job object - allows multiple jobs to run periodically in the same "daemon" cog.
  The max delay between jobs is about 27 seconds because it uses the cnt register for
  timestamps and keeping the timestamp math positive makes it easier.

  API:
    start
    addJob(fnctPtr, arg, timeTicks, repeatFlag)
      fnctPtr    - address of function pointer longs that point to batch job routine.
      arg        - argument to be passed to batch job routine
      timeTicks  - time to wait before routine is run.
                   In timer ticks. Max value is 2_147_483_647.
      repeatFlag - false - batch job is a one shot
                   true  - run batch job periodically
    cancelJob(jobId)

  Example usage:

OBJ
  mp:       "MethodPointer"
  batch:    "batch"
  debug:    "MultiCogSerialDebug" ' optional for debug messages to serial port

VAR
  long fnctPtr1[2]
  long fnctPtr2[2]

PUB Main
  .
  .
  .

  'Initialize the method pointer routines
  mp.Initialize  ' this must be run prior to calling batch.Start
  batch.Start
  if mp.SetMethodPtrEx(@fnctPtr1)
    tstJobX(0)
  if mp.SetMethodPtrEx(@fnctPtr2)
    tstJobY(0)
  .
  .
  .
  'add batch job to run every 10 seconds
  retStat := batch.addJob(@fnctPtr1, $DEADBA11, clkfreq*10, true)
  debug.cprintf(String("add repeating job to batch queue. jobNum %x\r"), retStat, false)

  'add batch job to run every 5 seconds
  retStat := batch.addJob(@fnctPtr2, $DEADBA11, clkfreq*5, false)
  debug.cprintf(String("add non-repeating job to batch queue. jobNum %x\r"), retStat, false)

PUB tstJobX(arg1)
  debug.cprintf(String("tstJob - arg1 = %x\r"), arg1, false)
  'do something useful

PUB tstJobY(arg1)
  debug.cprintf(String("tstJob - arg1 = %x\r"), arg1, false)
  'do something useful
  retStat := batch.addJob(@fnctPtr2, $DEADBA11, clkfreq*5, false)
  debug.cprintf(String("add non-repeating job to batch queue. jobNum %x\r"), retStat, false)
}
CON

'config this for the max number of simultaneous batch jobs that will be run
MAX_BATCH_JOBS = 8

'adjust this to size required
BATCH_DAEMON_STACK_SIZE = 128

'return status codes from addJob
OK                 = 0
TIME_TOO_LONG_FAIL = -1
JOB_QUEUE_FULL     = -2
NO_COGS_LEFT       = -3
NO_SEMAPHORES_LEFT = -4

{
  pseudo C def for our batch job control block

  enum BATCH_FLAG_IN_USE, BATCH_FLAG_REPEAT,
  struct batchJob
  {
    long fnctPtr[2];
    long timeWait;
    long timeStamp;
    long arg1;
    byte flags;
    byte  filler[3]; 'keep long word aligned
  }
}
batchFnctPtrOffset   = 0
batchTimeWaitOffset  = 8
batchTimeStampOffset = 12
batchArg1Offset      = 16
batchFlagsOffset     = 20
batchJobSize = 24

batchFlagJobInUse   = 0
batchFlagRepeat     = 1

MAX_POS_NUM = 2_147_483_647

'#define DEBUG


OBJ
  mp:       "MethodPointer"

VAR
  long batchStack[BATCH_DAEMON_STACK_SIZE]

DAT
  batchJobList  long 0[batchJobSize * MAX_BATCH_JOBS]
  batchQueSemId byte 0
  batchCogId    byte 0

PUB start : retStat
  retStat := -1
  batchQueSemId := locknew
  if (batchQueSemId == -1)
    retStat := NO_SEMAPHORES_LEFT
  else
    batchCogId := cogNew(batchQueDaemon(batchQueSemId, clkfreq/10), @batchStack[0])
    if (batchCogId == -1)
      retStat := NO_COGS_LEFT
    else
      retStat := 0

PUB addJob(fnctPtr, arg, timeTicks, repeatFlag) : batchJobIdx | ptr
  if (timeTicks < 0) OR (timeTicks > MAX_POS_NUM)
    batchJobIdx := TIME_TOO_LONG_FAIL
  else

    'look for an available slot
    repeat batchJobIdx from 0 to MAX_BATCH_JOBS -1

      repeat until not lockset(batchQueSemId)

      ptr := (@batchJobList + (batchJobIdx * batchJobSize))

      if (NOT(byte[ptr + batchFlagsOffset] & (|< batchFlagJobInUse)))

        if (repeatFlag == true)
          byte[ptr + batchFlagsOffset] |= (|< batchFlagRepeat)
        longmove((ptr + batchFnctPtrOffset), fnctPtr, 2)
        long[ptr + batchTimeWaitOffset]  := timeTicks
        long[ptr + batchArg1Offset]      := arg
        long[ptr + batchTimeStampOffset] := getPosCnt
        byte[ptr + batchFlagsOffset] |= (|< batchFlagJobInUse)
        lockclr(batchQueSemId)
        quit

      lockclr(batchQueSemId)

    'end repeat loop
  'endif


  if (batchJobIdx == MAX_BATCH_JOBS)
    batchJobIdx := JOB_QUEUE_FULL

PUB cancelJob(jobId) | ptr
  repeat until not lockset(batchQueSemId)

  ptr := (@batchJobList + (jobId * batchJobSize))
  if (jobId < MAX_BATCH_JOBS)
    byte[ptr + batchFlagsOffset] &= !(|< batchFlagJobInUse)

  lockclr(batchQueSemId)

'launched from start routine. Wake up every timerRes ticks and scan the entire list
'for routines that are ready to run.
PRI batchQueDaemon(bQlockId, timerRes) | ptr, batchJobIdx
  repeat
    waitcnt(cnt + timerRes)

    repeat batchJobIdx from 0 to MAX_BATCH_JOBS-1
      ptr := (@batchJobList + (batchJobIdx * batchJobSize))

      repeat until not lockset(batchQueSemId)
      'is there an active job in this slot?
      if (byte[ptr + batchFlagsOffset] & (|< batchFlagJobInUse))
        'is it time to run?
        if (timeStampExpired((ptr + batchTimeStampOffset), 1, long[ptr + batchTimeWaitOffset]))

          'is this a one-shot - not a periodically repeating job?
          if NOT (byte[ptr + batchFlagsOffset] & (|< batchFlagRepeat))
            'yes. free up the slot. do this before routine runs so that the
            'routine can manually resubmit the job if it needs to.
            byte[ptr + batchFlagsOffset] &= !(|< batchFlagJobInUse)

          lockclr(batchQueSemId)

          mp.CallMethod1(long[ptr + batchArg1Offset], (ptr + batchFnctPtrOffset))
          next

      lockclr(batchQueSemId)

    'end repeat batchIdx loop

  'end repeat forever loop

PRI getPosCnt : retCnt
  retCnt := cnt
  if retCnt < 0
    retCnt := MAX_POS_NUM - (||retCnt + 1)

PRI timeStampExpired(timeStampAddress, retryCount, timeoutVal) : retStat | timeStamp, upperLimit, saveCnt
  if (retryCount == 0)
    retStat := true
  else
    retStat := false

    saveCnt := getPosCnt

    timeStamp := long[timeStampAddress]

    if (saveCnt < timeStamp)
      if (((MAX_POS_NUM - timeStamp) + saveCnt) > timeoutVal)
        retStat := true 'timer expired
    elseif ((saveCnt - timeStamp) > timeoutVal)
      retStat := true 'timer expired

  if (retStat == true)
    'update the timestamp
    long[timeStampAddress] := getPosCnt
    if (retryCount > 0)


